import java.util.HashMap;
import com.cycling74.max.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.BasicConfigurator;
import com.rapplogic.xbee.api.AtCommand;
import com.rapplogic.xbee.api.AtCommandResponse;
import com.rapplogic.xbee.api.XBee;
import com.rapplogic.xbee.api.XBeeAddress16;
import com.rapplogic.xbee.api.XBeeAddress64;
import com.rapplogic.xbee.api.XBeeResponse;
import com.rapplogic.xbee.api.zigbee.*;
import com.rapplogic.xbee.api.*;

/**
 * @author CRX091081GB
 * 
 * TODO add Series 1 support (wpan discover as well as 
 *
 */

public class XBee4Max extends MaxObject {
	
	//version tracking
	private final float version = 1.04f;
	private final String status = "rc";
	
	//API object allows us to access our coordinator
	XBee xbee;
	
	//Private class sets up threads for listening
	//each has a boolean we turn off when we want to stop
	private volatile Listener listener;
	volatile boolean open;
	
	String port = null;
	int baud;
	boolean debug = false;
	boolean acknowledge = false;
	
	private final static Logger log = Logger.getRootLogger();
    
    private HashMap<String, XBeeAddress64> names2addresses64;
    private HashMap<String, XBeeAddress16> names2addresses16;
    private HashMap<XBeeAddress64, String> addresses2names;
    
	//constructor
	public XBee4Max(Atom[] args) throws Exception {
		//post boiler plate
		post("--------------------------------------------");
		post("mxj XBee4Max, " + status + "-" + version + "\n");
		post("Robin Price 2011\n");
		post("--------------------------------------------");
		
		//This is useful for finding out the right place to drop the libRxTxSerial files
		//String path = System.getProperty("java.library.path");
		//post(path+"\n");
		
		//class attributes
		declareAttribute("port");
		declareAttribute("baud");
		declareAttribute("debug");
		declareAttribute("acknowledge");
		
		declareOutlets(new int[]{DataTypes.LIST, DataTypes.MESSAGE, DataTypes.MESSAGE, DataTypes.MESSAGE, DataTypes.MESSAGE});
		
		setInletAssist(new String[]{"open, close and send followed by a valid name and some data"});
		setOutletAssist(new String[]{"Name of transmitting XBee and list of received data bytes", "bang when done sending", "bang on transmission success (acknowledge mode only)", "bang on transmission failure (acknowledge mode only)", "Name of XBee on discovery"});
		
		
		//turn logging off
		BasicConfigurator.configure();
		log.setLevel(Level.OFF);
		
		//hashtables for mapping xbee names to 64 and 16 bit addresses and 64 bit addresses back to names
		names2addresses64 = new HashMap<String, XBeeAddress64>();
		names2addresses16 = new HashMap<String, XBeeAddress16>();
		addresses2names = new HashMap<XBeeAddress64, String>();
	}
	 
	//start the connection with the serial port
	public void open() {
		if (!open && port != null && (baud == 1200 || baud == 2400 || baud == 9600
				|| baud == 19200 || baud == 38400 || baud == 57600
				|| baud == 115200 || baud == 230400)) {
			try {
				xbee = new XBee();
				xbee.open(port, baud);
				listener = new Listener();
				listener.start();
				open = true;
				//send out node discovery command
				xbee.sendAsynchronous(new AtCommand("ND"));
				//send 'clear' from outlet 5 to clear any colls or umenus the user is using to keep track of connected xbees
				outlet(4, "clear");
				post("XBee4Max: Port opened\n");
			} catch (Exception e) {
				post("XBee4Max: ERROR\n");
				post("XBee4Max: Cannot open port\n");
				//post("ZBee4Max: " + e.getCause().getMessage() + "\n");
				open = false;
			}
		} else {
			if (xbee == null) {
				post("XBee4Max: Cannot open without valid port and baud rate.\n");
			}
		}
	}
	
	//destructor
	public void notifyDeleted() {
		this.close();
	}
	
	//stop the party
	public void close() {
		//stop listener thread by setting open false
		this.open = false;
		if (xbee!=null)
		if (xbee.isConnected()) {
			xbee.close();
			xbee = null;
			post("XBee4Max: Port closed\n");
		}
	}
	//this seems to fail so scrap it
	/*
	public void reset() {
		this.close();
		this.open();
	}
	*/
	
	public void send(final Atom[] args) {
		if (open) { //only attempt to send if we've got a serial port and an xbee on the end
			Thread t = new Thread() {//run the send request in it's own thread so the audio won't get mucked up
				public void run() {
					boolean legal = true; //assume arguments supplied with send request are correct till proven wrong
					if (args.length > 1) { //check we have args
						for (int i = 1; i < args.length; i++) {//iterate thru args
							Atom a = args[i];
							if (a.isInt() == false) {//args need to be ints
								legal = false;
								post("XBee4Max: All send data must be ints within the range 0 - 255\n");
							} else {
								if (a.getInt() > 255 || a.getInt() < 0) { //args need to be in range for a byte
									legal = false;
									post("XBee4Max: All send data must be ints within the range 0 - 255\n");
								}
							}
						}//finished iterating
					} else {//we had no args
						legal = false; //no arguments supplied, set legal false and next if statement will be skipped
						post("XBee4Max: send command requires valid data to send\n");
					}
					//if the data we've been asked to send is good and we have an address for the name supplied or we're sending to ALL then send
					if (legal && (names2addresses64.containsKey(args[0].getString()) || args[0].getString().equalsIgnoreCase("all"))) {
						int[] payload = new int[args.length - 1];
						for (int i = 1; i < args.length; i++) {
							payload[i - 1] = args[i].toInt(); //format payload
						}
						if (args[0].toString().equalsIgnoreCase("all")) { //take care of sending to all, note that broadcast packets CANNOT BE SENT with acknowledgement on delivery.
							ZNetTxRequest request = new ZNetTxRequest(ZNetTxRequest.NO_RESPONSE_FRAME_ID, XBeeAddress64.BROADCAST, XBeeAddress16.BROADCAST, ZNetTxRequest.DEFAULT_BROADCAST_RADIUS, ZNetTxRequest.Option.BROADCAST, payload);
							try {
								xbee.sendAsynchronous(request); //send it 
							} catch (Exception e) {
								post("XBee4Max: ERROR\n");
								e.printStackTrace();
							}
						} else { //else it's not a broadcast packet, send unicast with acknowledge off or one depending on attribute 
							//format transmission request
							if (debug) {
								post("Sending to 64 bit address: " + names2addresses64.get(args[0].toString()).toString() + "\n");
								post("Sending to 16 bit address: " + names2addresses16.get(args[0].toString()).toString() + "\n");
							}
							
							try {
								if (acknowledge) {//send and wait for acknowledgment
									ZNetTxRequest request = new ZNetTxRequest(ZNetTxRequest.DEFAULT_FRAME_ID, names2addresses64.get(args[0].toString()), names2addresses16.get(args[0].toString()), ZNetTxRequest.DEFAULT_BROADCAST_RADIUS, ZNetTxRequest.Option.UNICAST, payload);
									ZNetTxStatusResponse response = (ZNetTxStatusResponse) xbee.sendSynchronous(request, 10000); //send, wait for upto 10 seconds for response
									
									request.setFrameId(xbee.getNextFrameId());
									if (response.getDeliveryStatus() == ZNetTxStatusResponse.DeliveryStatus.SUCCESS) {
										// the packet was successfully delivered
										outlet(2, "bang");							
									} else {
										outlet(3, "bang");
										// packet failed.  log error
										// it's easy to create this error by unplugging/powering off your remote xbee.  when doing so I get: packet failed due to error: ADDRESS_NOT_FOUND  
										post("XBee4Max: Packet transmission failed due to error: " + response.getDeliveryStatus());
									}
								} else {//send without waiting for acknowledgment
									ZNetTxRequest request = new ZNetTxRequest(ZNetTxRequest.NO_RESPONSE_FRAME_ID, names2addresses64.get(args[0].toString()), names2addresses16.get(args[0].toString()), ZNetTxRequest.DEFAULT_BROADCAST_RADIUS, ZNetTxRequest.Option.UNICAST, payload);
									xbee.sendAsynchronous(request); //send
								}
								outlet(1, "bang"); //we're done trying to send
							} catch (Exception e) {
								post("XBee4Max: ERROR\n");
								e.printStackTrace();
							}
						}//done sending
					} else {//illegal name
						if (legal) {
							post("XBee4Max: No XBee with adress " + args[0].toString() + " has been logged\n");
						}
					}
				}
			};
			t.start();
		} else {
			post("XBee4Max: Can't send, port not open\n");
		}
	}

	private class Listener extends Thread {
		public void run() {
			while(open) {
				try {
					XBeeResponse response = xbee.getResponse();
					if (debug) {
						post("XBee4Max: Received response frame" + response.toString());
					}
					if (response.getApiId() == ApiId.AT_RESPONSE) {
						ZBNodeDiscover nd = ZBNodeDiscover.parse((AtCommandResponse)response); //If you  want to add Series 1 support then this part will need changing to include WPan
						if (debug) {
							post("Node discover response is: " + nd);
						}
						//send out names of any XBee's that were already on when we turned on the coordinator
						//these messages will be triggered by the nd command in the open() code
						outlet(4, nd.getNodeIdentifier());
						names2addresses64.put(nd.getNodeIdentifier(), nd.getNodeAddress64());
						names2addresses16.put(nd.getNodeIdentifier(), nd.getNodeAddress16());
						addresses2names.put(nd.getNodeAddress64(), nd.getNodeIdentifier());
					}
					if (response.getApiId() == ApiId.ZNET_IO_NODE_IDENTIFIER_RESPONSE) {
						ZNetNodeIdentificationResponse rx = (ZNetNodeIdentificationResponse) response;
						if (debug) {
							post("XBee4Max: " + rx.getNodeIdentifier() + " has notified with 64 bit address " + rx.getRemoteAddress64().toString() + " and 16 bit address " + rx.getRemoteAddress16());
						}
						//send out name of new XBee from last outlet
						outlet(4, rx.getNodeIdentifier());
						names2addresses64.put(rx.getNodeIdentifier(), rx.getRemoteAddress64());
						names2addresses16.put(rx.getNodeIdentifier(), rx.getRemoteAddress16());
						addresses2names.put(rx.getRemoteAddress64(), rx.getNodeIdentifier());
						
					}
					if (response.getApiId() == ApiId.ZNET_RX_RESPONSE) {
						ZNetRxResponse rx = (ZNetRxResponse) response;
						if (addresses2names.containsKey(rx.getRemoteAddress64())) {
							//format output array with name of transmitting xbee at front
							Atom[] Output = new Atom[rx.getData().length + 1];
							Output[0] = Atom.newAtom(addresses2names.get(rx.getRemoteAddress64()));
							for (int i = 0; i < rx.getData().length; i++) {
								Output[i + 1] = Atom.newAtom(rx.getData()[i]);
							}
							outlet(0, Output);
							//keep 16 bit tables up to date
							names2addresses16.put(addresses2names.get(rx.getRemoteAddress64()), rx.getRemoteAddress16());
						} else {
							//got data from an xbee with out a name!
							//warn
							post("XBee4Max: Received data from unknown XBee!");
							//reissue node discovery command so we can get their name and address
							xbee.sendAsynchronous(new AtCommand("ND"));
						}
					}
				} catch (Exception e) {
					post("XBee4Max: ERROR");
					e.printStackTrace();
				}
			}
		}
	}
}